"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin


class MuonPlugin(Plugin):
    """
    A plugin for labeling muons in charge and light data.

    Muons are MIPs which deposit energy through MIP ionization (i.e. the
    standard dE/dx mechanism).  This plugin grabs all muons in an event
    and labels their energy deposits according to the following scheme:

        (1) hits coming directly from muon ionization are labeled as
                topology = track
                physics_micro = mip_ionization
                physics_macro = mip
        (2) delta electrons which are daughters of muons are labeled as
                topology = track
                physics_micro = electron_ionization
                physics_macro = delta_electron
        (3) michel electrons which are daughters of muons are labeled as
                topology = track
                physics_micro = electron_ionization
                physics_macro = michel_electron

    Apart from these labels, track beginning and ending points are also
    labelled.  Vertices are placed at points where Michel electrons are
    generated.

    StandardRecord objects are also generated.
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(MuonPlugin, self).__init__(config)

        self.input_products = ['daughters', 'track_id_hit_map']
        self.output_product = None

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        charge = flow_file['charge/calib_final_hits/data']

        muons = abs(trajectories['pdg_id']) == 13



"""
"""
from mpi4py import MPI
import h5py
import os
import importlib.util
import traceback
import sys
import inspect
import glob
from tqdm import tqdm
import torch
import numpy as np
from datetime import datetime
from matplotlib import pyplot as plt
from importlib.metadata import version

from arrakis_nd.utils.logger import Logger
from arrakis_nd.utils.utils import (
    profiler,
    timing_manager,
    memory_manager
)
from arrakis_nd.plugins.plugin import Plugin


class Arrakis:
    """
    Main Arrakis class for running jobs. Arrakis is designed
    to work with MPI and H5.  The user must run Arrakis using mpirun
    with at least two processes (one master and N-1 workers).
    
    
    """
    @profiler
    def __init__(
        self,
        config: dict = {},
        meta:   dict = {},
    ):
        """_summary_

        Args:
            config (dict): config file for running Arrakis.
            meta (dict): dictionary of meta information to
            be shared across all nodes.
        """

        """Get mpi communication parameters"""
        try:
            self.comm = MPI.COMM_WORLD
            self.rank = self.comm.Get_rank()
            self.size = self.comm.Get_size()
        except Exception as e:
            raise RuntimeError(f"unable to obtain MPI parameters: {e}")

        """Set input parameters and set up loggers"""
        try:
            if self.rank == 0:
                self.logger = Logger(f"master: {self.rank}", output="both")
            else:
                self.logger = Logger(f"worker: {self.rank}", output="both")
        except Exception as e:
            raise RuntimeError(f"unable to set up logging system: {e}")

        self.config = config
        self.meta = meta

        """Setting error status for this node"""
        self.error_status = None

        """Parse config"""
        try:
            self.parse_config()
        except Exception as e:
            self.error_status = e
        self.barrier()

        """Construct plugins"""
        self.plugins = {}
        try:
            self.construct_plugins()
        except Exception as e:
            self.error_status = e
        self.barrier()

        self.flow_files = []
        """Distributed events and indices for various arrays"""
        self.distributed_events = {}
        self.distributed_interactions_indices = {}
        self.distributed_segments_indices = {}
        self.distributed_stack_indices = {}
        self.distributed_trajectories_indices = {}
        self.distributed_charge_indices = {}
        self.distributed_light_indices = {}

    @profiler
    def barrier(self):
        errors = self.comm.allgather(self.error_status)
        if any(errors):
            if self.rank == 0:
                errors_count = sum(1 for error in errors if error is not None)
                errors_indices = [index for index, error in enumerate(errors) if error is not None]
                self.logger.critical(f"{errors_count} errors encountered in Arrakis program")
                for index in errors_indices:
                    self.logger.critical(f"error encountered in worker {index}: ")
                    self.logger.critical(errors[index])
                self.comm.Abort(1)
        else:
            self.comm.Barrier()

    @profiler
    def parse_config(self):
        """
        Set up config parameters from input config file
        """
        if self.rank == 0:
            try:
                """Startup main arrakis program"""
                self.logger.info(f'############################ ARRAKIS  v. [{version("arrakis-nd")}] ############################')
                
                """Check for main arrakis_nd parameters"""
                if "arrakis_nd" not in self.config.keys():
                    self.logger.error("arrakis_nd section not in config!")
                arrakis_nd_config = self.config["arrakis_nd"]

                """Try to grab system info and display to the logger"""
                system_info = self.logger.get_system_info()
                time = datetime.now()
                now = f"{time.hour}:{time.minute}:{time.second} [{time.day}/{time.month}/{time.year}]"
                self.logger.info(f'system_info - local time: {now}')
                for key, value in system_info.items():
                    self.logger.info(f"system_info - {key}: {value}")

                """Set the verbosity of the Arrakis program"""
                if "verbose" in arrakis_nd_config:
                    if not isinstance(arrakis_nd_config["verbose"], bool):
                        self.logger.error(
                            f'"arrakis_nd:verbose" must be of type bool, but got {type(arrakis_nd_config["verbose"])}!'
                        )
                    self.meta["verbose"] = arrakis_nd_config["verbose"]
                else:
                    self.meta["verbose"] = False

                """See if any CUDA devices are available on the system"""
                if torch.cuda.is_available():
                    self.logger.info("CUDA is available with devices:")
                    for ii in range(torch.cuda.device_count()):
                        device_properties = torch.cuda.get_device_properties(ii)
                        self.logger.info(f" -- device: {ii}")
                        self.logger.info(f" {' ':<{5}} name: {device_properties.name}")
                        self.logger.info(f" {' ':<{5}} compute: {device_properties.major}.{device_properties.minor}")
                        self.logger.info(f" {' ':<{5}} memory: {round(device_properties.total_memory / (1024**3))} GB")

                """Check for GPU devices and configure them"""
                if "gpu" not in arrakis_nd_config.keys():
                    self.logger.info('"arrakis_nd:gpu" not specified in config')
                    gpu = None
                else:
                    gpu = arrakis_nd_config["gpu"]
                if "gpu_device" not in arrakis_nd_config.keys():
                    self.logger.info('"arrakis_nd:gpu_device" not specified in config')
                    gpu_device = None
                else:
                    gpu_device = arrakis_nd_config["gpu_device"]

                """Set the GPU parameters"""
                if gpu:
                    if torch.cuda.is_available():
                        if gpu_device >= torch.cuda.device_count() or gpu_device < 0:
                            self.logger.warn(
                                f"desired gpu_device '{gpu_device}' not available, using device '0'"
                            )
                            gpu_device = 0
                        self.meta["device"] = torch.device(f"cuda:{gpu_device}")
                        self.logger.info(
                            f"CUDA is available, using device {gpu_device}"
                            + f": {torch.cuda.get_device_name(gpu_device)}"
                        )
                    else:
                        gpu is False
                        self.logger.warn("CUDA not available! Using the cpu")
                        self.meta["device"] = torch.device("cpu")
                else:
                    self.logger.info("using cpu as device")
                    self.meta["device"] = torch.device("cpu")
                self.meta["gpu"] = gpu

                """Send new device meta data to workers"""
                for ii in range(1, self.comm.size):
                    self.comm.send(self.meta, dest=ii, tag=0)
            except Exception as e:
                self.error_status = e
            self.barrier()
        else:
            self.barrier()
            try:
                self.meta = self.comm.recv(source=0, tag=0)
            except Exception as e:
                self.error_status = e

    @profiler
    def construct_plugins(self):
        """
        Construct all plugin objects.  First, we look for all associated
        classes which inherit from plugins.  These include both the standard
        plugins which are merged into the plugins/ folder, as well as
        any which are specified by the user at run time.
        """
        """First collect all available plugins"""
        self.collect_plugins()

        """Check that plugins section exists and has some actual plugins in it"""
        if 'plugins' not in self.config.keys():
            self.logger.error('no plugins section specified in config!')
        plugins_config = self.config['plugins']
        if plugins_config is None:
            self.logger.error('no plugins specified in config!')

        """Iterate over specified plugins and check that they exist"""
        for plugin, config in plugins_config.items():
            if plugin not in self.available_plugins.keys():
                self.logger.error(f'specified plugin "{plugin}" not available!')
            try:
                self.plugins[plugin] = self.available_plugins[plugin](config)
            except Exception as e:
                self.logger.error(f'failed to construct plugin {plugin} with config {config}: {e}')

        """Check that intermediate products are set up correctly"""
        running_output_products = []
        for ii, plugin_name in enumerate(self.plugins.keys()):
            if ii == 0:
                if self.plugins[plugin_name].input_products is not None:
                    self.logger.error(
                        f'first specfied plugin "{plugin_name}" should have input_products == None, \
                        but has input_products == "{self.plugins[plugin_name].input_products}"!'
                    )
            else:
                if (
                    (self.plugins[plugin_name].input_products is None) or
                    (self.plugins[plugin_name].input_products == [])
                ):
                    pass
                else:
                    if isinstance(self.plugins[plugin_name].input_products, str):
                        if self.plugins[plugin_name].input_products not in running_output_products:
                            self.logger.error(
                                f'specified plugin "{plugin_name}" at location "{ii}" expects input_products \
                                "{self.plugins[plugin_name].input_products}" but current products only contain \
                                {running_output_products}!'
                            )
                    elif isinstance(self.plugins[plugin_name].input_products, list):
                        for input_product in self.plugins[plugin_name].input_products:
                            if input_product not in running_output_products:
                                self.logger.error(
                                    f'specified plugin "{plugin_name}" at location "{ii}" expects input_products \
                                    "{self.plugins[plugin_name].input_products}" but current products only contain \
                                    {running_output_products}!'
                                )
            running_output_products.append(self.plugins[plugin_name].output_product)

    @profiler
    def collect_plugins(self):
        """
        This function collects all available plugins from the ArrakisND
        plugins folder, as well as any plugin locations specified in the
        config file.
        """
        self.available_plugins = {}
        self.plugin_files = [
            os.path.dirname(__file__) + '/../plugins/' + file
            for file in os.listdir(path=os.path.dirname(__file__) + '/../plugins/')
        ]
        if 'plugin_locations' in self.config.keys():
            if (
                (not isinstance(self.config['plugin_locations'], list)) and
                (self.config['plugin_locations'] is not None)
            ):
                self.logger.error(
                    f"plugin_locations specified in config is of type {type(self.config['plugin_locations'])}, \
                    but must be a list or none!"
                )
            self.plugin_files.extend(self.config['plugin_locations'])

        """Iterate over found plugin files and attempt to load them"""
        for plugin_file in self.plugin_files:
            if (
                ("__init__.py" in plugin_file) or
                ("__pycache__.py" in plugin_file) or
                ("__pycache__" in plugin_file) or
                (".py" not in plugin_file)
            ):
                continue
            try:
                self.load_plugin_function(plugin_file)
            except Exception as e:
                self.logger.warn(f'problem loading plugin from file: {plugin_file}: {e}')

    @profiler
    def load_plugin_function(
        self,
        plugin_file: str
    ):
        """
        This function attempts to load a plugin which inherits from
        the Plugin class.

        Args:
            plugin_file (str): _description_
        """
        spec = importlib.util.spec_from_file_location(
            f'{plugin_file.removesuffix(".py")}.name',
            plugin_file
        )
        custom_plugin_file = importlib.util.module_from_spec(spec)
        sys.modules[f'{plugin_file.removesuffix(".py")}.name'] = custom_plugin_file
        spec.loader.exec_module(custom_plugin_file)
        for name, obj in inspect.getmembers(sys.modules[f'{plugin_file.removesuffix(".py")}.name']):
            if inspect.isclass(obj):
                custom_class = getattr(custom_plugin_file, name)
                if issubclass(custom_class, Plugin):
                    self.available_plugins[name] = custom_class

    @profiler
    def set_up_input_files(self):
        """
        Iterate over flow_folder directory and determine
        if listed input files exist, or if "all" is selected for
        input files, construct the list of all .h5 files in the
        flow_folder.
        """
        if self.rank == 0:
            try:
                arrakis_dict = self.config['arrakis_nd']
                
                """Check for parameters"""
                if 'flow_folder' not in arrakis_dict.keys():
                    self.logger.error(f'flow_folder not specified in config!')
                if 'flow_files' not in arrakis_dict.keys():
                    self.logger.error(f'flow_files not specified in config!')
                    
                flow_folder = arrakis_dict['flow_folder']
                flow_files = arrakis_dict["flow_files"]
                
                """Check for arrakis folder"""
                if 'arrakis_folder' not in arrakis_dict.keys():
                    self.logger.warn(f'arrakis_folder not specified in config! setting to "/local_scratch"')
                    arrakis_dict['arrakis_folder'] = '/local_scratch'
                arrakis_folder = arrakis_dict['arrakis_folder'].replace('flow', 'arrakis').replace('FLOW', 'ARRAKIS')
                                
                """Check that flow folder exists"""
                if not os.path.isdir(flow_folder):
                    self.logger.error(f'specified flow_folder {flow_folder} does not exist!')
                    
                """Check that flow folder has a '/' at the end"""
                if flow_folder[-1] != '/':
                    flow_folder += '/'
                    
                """Check that arrakis folder has a '/' at the end"""
                if arrakis_folder[-1] != '/':
                    arrakis_folder += '/'
                    
                """Check that arrakis folder exists"""
                if not os.path.isdir(arrakis_folder):
                    os.makedirs(arrakis_folder)

                if isinstance(arrakis_dict["flow_files"], list):
                    """
                    If the flow_files parameter is a list, look through
                    the list and make sure each specified file actually exists
                    in the flow_folder.
                    """
                    flow_files = [
                        input_file for input_file in arrakis_dict["flow_files"]
                        if input_file not in arrakis_dict["skip_files"]
                    ]
                    for flow_file in flow_files:
                        if not os.path.isfile(flow_file):
                            self.logger.error(
                                f"specified file {flow_file} does not exist in directory {flow_folder}!"
                            )
                elif isinstance(arrakis_dict["flow_files"], str):
                    """
                    If the flow_files parameter is a string, check if its
                    the phrase 'all', and if so, recursively grab all h5
                    files in the flow_folder.

                    Otherwise, assume that the flow_files parameter is a
                    file extension, and search recursively for all files
                    with that extension.
                    """
                    if arrakis_dict["flow_files"] == "all":
                        self.logger.info(
                            f"searching {flow_folder} recursively for all .h5 FLOW files."
                        )
                        flow_files = [
                            os.path.basename(input_file) for input_file in glob.glob(
                                f"{flow_folder}*.h5", recursive=True
                            )
                            if 'FLOW' in input_file and input_file not in arrakis_dict["skip_files"]
                        ]
                    else:
                        try:
                            self.logger.info(
                                f'searching {flow_folder} recursively for all {arrakis_dict["flow_files"]} files.'
                            )
                            flow_files = [
                                os.path.basename(input_file) for input_file in glob.glob(
                                    f'{flow_folder}/{arrakis_dict["flow_files"]}',
                                    recursive=True,
                                )
                                if input_file not in arrakis_dict["skip_files"]
                            ]
                        except Exception as exception:
                            self.logger.error(
                                f'specified "flow_files" parameter: {arrakis_dict["flow_files"]} incompatible!'
                                + f" exception: {exception}"
                            )
                else:
                    self.logger.error(
                        f'specified "flow_files" parameter: {arrakis_dict["flow_files"]} incompatible!'
                    )
                self.flow_folder = flow_folder
                self.flow_files = flow_files
                self.arrakis_folder = arrakis_folder
                self.logger.info(f'setting arrakis_folder to {self.arrakis_folder}.')
                for ii in range(1, self.comm.size):
                    self.comm.send(self.flow_folder, dest=ii, tag=80)
                    self.comm.send(self.flow_files, dest=ii, tag=81)
                    self.comm.send(self.arrakis_folder, dest=ii, tag=82)
                self.logger.info(f'found {len(self.flow_files)} flow files for processing.')
            except Exception as e:
                self.error_status = e
            self.barrier()
        else:
            self.barrier()
            try:
                self.flow_folder = self.comm.recv(source=0, tag=80)
                self.flow_files = self.comm.recv(source=0, tag=81)
                self.arrakis_folder = self.comm.recv(source=0, tag=82)
            except Exception as e:
                self.error_status = e

    @profiler
    def run_begin_of_file(
        self,
        file_name: str
    ):
        """
        Plugins to be run at the beginning of the file,
        which act on the entire file.

        Args:
            file_name (_str_): name of the input flow file
        """
        pass

    @profiler
    def set_up_output_arrays(
        self,
        file_name: str
    ):
        """
        Construct output arrays for various data objects.
        This will take the flow file and create a new .h5 file
        replacing the words "FLOW" with "ARRAKIS" and "flow"
        with "arrakis".

        Args:
            file_name (str): name of the input flow file
        """

        """Determine the arrakis file name"""
        arrakis_file_name = file_name.replace('FLOW', 'ARRAKIS').replace('flow', 'arrakis')

        """Open the flow file and determine the output array shapes"""
        with h5py.File(self.flow_folder + file_name, 'r') as flow_file, \
            h5py.File(self.arrakis_folder + arrakis_file_name, 'a') as arrakis_file:
            """
            First we make the labels for the charge dataset.  These consist of six
            main labels that we wish to generate with various plugins:
                (1) topology - a descriptor of basic shapes
                (2) physics_micro - a descriptor of local physics
                (3) physics_macro - a descriptor of larger scale physics objects
                (4) unique_topology - unique labels for instances of topology
                (5) unique_physics_micro - unique labels for instances of physics_micro
                (6) unique_physics_macro - unique labels for instances of physics_macro
                (7) vertex - binary variable denoting whether a vertex is at this hit
                (8) tracklette_begin - binary variable denoting whether a track beginning is at this hit
                (9) tracklette_end - binary variable denoting whether a track end is at this hit
                (10) fragment_begin - binary variable denoting whether a fragment beginning is at this hit
                (11) fragment_end - binary variable denoting whether a fragment end is at this hit
                (12) shower_begin - binary variable denoting whether a shower beginning is at this hit

            These labels are assigned to each reconstructed charge hit and
            written in the corresponding ARRAKIS file.
            """
            charge_name = 'charge/calib_final_hits/data'

            num_charge = len(flow_file[charge_name]['x'])

            new_charge_data_type = np.dtype([
                ('event_id', 'i4'),
                ('topology', 'i4'),
                ('physics_micro', 'i4'),
                ('physics_macro', 'i4'),
                ('unique_topology', 'i4'),
                ('unique_physics_micro', 'i4'),
                ('unique_physics_macro', 'i4'),
                ('vertex', 'bool'),
                ('tracklette_begin', 'bool'),
                ('tracklette_end', 'bool'),
                ('fragment_begin', 'bool'),
                ('fragment_end', 'bool'),
                ('shower_begin', 'bool')
            ])

            new_charge_data = np.full(num_charge, -1, dtype=new_charge_data_type)

            if charge_name in arrakis_file:
                del arrakis_file[charge_name]

            arrakis_file.create_dataset(charge_name, data=new_charge_data)

            """
            Second, we set up arrays for CAF-like reco objects which consist
            of the following types:
                (1) tracklettes - pieces of contiguous track like objects.
                    (a) start (x,y,z) - track start point
                    (b) end (x,y,z) - track end point
                    (c) dir (u_x,u_y,u_z) - estimate of track direction taken from start point
                    (d) enddir (u_x,u_y,u_z) - estimate of track direction takend from end point
                    (e) Evis - visible energy in voxels corresponding to this track
                    (f) qual - reco specific quality metric
                    (g) len_gcm2 - track length in g/cm2
                    (h) len_cm - track length in centimeter
                    (i) E - track energy estimate in MeV
                    (j) truth - associated true particle in flow file (if relevant) (i.e. track_id)
                    (k) truthOverlap - fractional overlap between this track and true particle
                (2) tracks - collections of tracklettes which define a complete track.
                (3) fragments - pieces of contiguous shower like objects.
                    (a) start (x,y,z) - shower start point
                    (b) dir (u_x,u_y,u_z) - shower direction
                    (c) Evis - visible energy in voxels corresponding to this shower
                    (d) qual - reco specific quality metric
                    (e) len_gcm2 - track length in g/cm2
                    (f) len_cm - track length in centimeter
                    (g) E - track energy estimate in MeV
                    (h) truth - associated true particle in flow file (if relevant) (i.e. track_id)
                    (i) truthOverlap - fractional overlap between this shower and true particle
                (4) showers - collections of fragments which define a complete shower.
                (5) blips - collections of points associated to blip-like activity.
                    (a) start (x,y,z) - start position of this blip object
                    (b) Evis - visible energy in voxels corresponding to this blip
                    (c) E - reconstructed energy (GeV)
                    (d) bliphyp - hypothesis for this blip's identity
                    (e) truth - associated true particle in flow file (if relevant) (i.e track_id)
                    (f) truthOverlap - fractional overlap between this blip and true particle
                (6) particles - particles within an event that are associated to tracks/showers/blips.
                    (a) primary - is this reco particle a primary one (i.e. eminates directly from vertex)?
                    (b) pdg - pdg code inferred for this particle
                    (c) tgtA - atomic number of nucleus this particle was reconstructed in
                    (d) score - PID score for this particle
                    (e) E - reconstructed energy (GeV)
                    (f) E_method - method used to determine energy for the particle
                    (g) p (p_x,p_y,p_z) - reconstructed momentum for this particle
                    (h) start (x,y,z) - reconstructed start point of this particle
                    (i) end (x,y,z) - reconstructed end point of this particle
                    (j) contained - contained in LAr TPC?
                    (k) truth - associated true particle in flow file (if relevant) (i.e. track_id)
                    (l) truthOverlap - fractional overlap between this reco particle and true particle
                (7) interactions - collections of tracks/showers/blips which define interactions of interest.
                    (a) vtx (x,y,z) - reconstructed vertex location
                    (b) dir (u_x,u_y,u_z) - hypothesis for this interaction's parent particle direction
                        (b.i) lngtrk - direction using longest track
                        (b.ii) heshw - direction using highest energy shower
                    (c) nuhyp - hypothesis for this interaction's neutrino identity
                        (c.i) isnubar
                        (c.ii) nue
                        (c.iii) numu
                        (c.iv) nutau
                        (c.v) nc
                        (c.vi) protons0
                        (c.vii) protons1
                        (c.viii) protons2
                        (c.ix) protonsN
                        (c.x) chgpi0
                        (c.xi) chgpi1
                        (c.xii) chgpi2
                        (c.xiii) chgpiN
                        (c.xiv) pizero0
                        (c.xv) pizero1
                        (c.xvi) pizero2
                        (c.xvii) pizeroN
                        (c.xviii) neutron0
                        (c.xix) neutron1
                        (c.xx) neutron2
                        (c.xxi) neutronN
                    (d) Enu - hypothesis for this interaction's neutrino energy
                    (e) part - collections of reconstructed particles
                    (f) truth - indicies of true_interactions in flow file (if relevant)
                    (g) truthOverlap - fractional overlap between this reco interaction and each true interaction
            """

            """Construct track data types"""
            tracklette_name = 'standard_record/tracklette'
            track_name = 'standard_record/track'
            tracklette_data_type = np.dtype([
                ('event_id', 'i4'),
                ('tracklette_id', 'i4'),
                ('start', 'f4', (1, 3)),
                ('end', 'f4', (1, 3)),
                ('dir', 'f4', (1, 3)),
                ('enddir', 'f4', (1, 3)),
                ('Evis', 'f4'),
                ('qual', 'f4'),
                ('len_gcm2', 'f4'),
                ('len_cm', 'f4'),
                ('E', 'f4'),
                ('truth', 'i4', (1, 20)),
                ('truthOverlap', 'f4', (1, 20)),
            ])
            track_data_type = np.dtype([
                ('event_id', 'i4'),
                ('track_id', 'i4'),
                ('tracklette_ids', 'i4', (1, 20)),
                ('start', 'f4', (1, 3)),
                ('end', 'f4', (1, 3)),
                ('dir', 'f4', (1, 3)),
                ('enddir', 'f4', (1, 3)),
                ('Evis', 'f4'),
                ('qual', 'f4'),
                ('len_gcm2', 'f4'),
                ('len_cm', 'f4'),
                ('E', 'f4'),
                ('truth', 'i4', (1, 20)),
                ('truthOverlap', 'f4', (1, 20)),
            ])
            if tracklette_name in arrakis_file:
                del arrakis_file[tracklette_name]
            if track_name in arrakis_file:
                del arrakis_file[track_name]

            arrakis_file.create_dataset(tracklette_name, shape=(0,), maxshape=(None,), dtype=tracklette_data_type)
            arrakis_file.create_dataset(track_name, shape=(0,), maxshape=(None,), dtype=track_data_type)

            """Construct shower data types"""
            fragment_name = 'standard_record/fragment'
            shower_name = 'standard_record/shower'
            fragment_data_type = np.dtype([
                ('event_id', 'i4'),
                ('fragment_id', 'i4'),
                ('start', 'f4', (1, 3)),
                ('dir', 'f4', (1, 3)),
                ('Evis', 'f4'),
                ('qual', 'f4'),
                ('len_gcm2', 'f4'),
                ('len_cm', 'f4'),
                ('E', 'f4'),
                ('truth', 'i4', (1, 20)),
                ('truthOverlap', 'f4', (1, 20)),
            ])
            shower_data_type = np.dtype([
                ('event_id', 'i4'),
                ('shower_id', 'i4'),
                ('fragment_ids', 'i4', (1, 20)),
                ('start', 'f4', (1, 3)),
                ('dir', 'f4', (1, 3)),
                ('Evis', 'f4'),
                ('qual', 'f4'),
                ('len_gcm2', 'f4'),
                ('len_cm', 'f4'),
                ('E', 'f4'),
                ('truth', 'i4', (1, 20)),
                ('truthOverlap', 'f4', (1, 20)),
            ])
            if fragment_name in arrakis_file:
                del arrakis_file[fragment_name]
            if shower_name in arrakis_file:
                del arrakis_file[shower_name]

            arrakis_file.create_dataset(fragment_name, shape=(0,), maxshape=(None,), dtype=fragment_data_type)
            arrakis_file.create_dataset(shower_name, shape=(0,), maxshape=(None,), dtype=shower_data_type)

            """Construct blip data types"""
            blip_name = 'standard_record/blip'
            blip_data_type = np.dtype([
                ('event_id', 'i4'),
                ('blip_id', 'i4'),
                ('start', 'f4', (1, 3)),
                ('Evis', 'f4'),
                ('E', 'f4'),
                ('bliphyp', 'i4'),
                ('truth', 'i4', (1, 20)),
                ('truthOverlap', 'f4', (1, 20)),
            ])
            if blip_name in arrakis_file:
                del arrakis_file[blip_name]

            arrakis_file.create_dataset(blip_name, shape=(0,), maxshape=(None,), dtype=blip_data_type)

            """Construct particle data types"""
            particle_name = 'standard_record/particle'
            particle_data_type = np.dtype([
                ('event_id', 'i4'),
                ('particle_id', 'i4'),
                ('primary', 'bool'),
                ('pdg', 'i4'),
                ('tgtA', 'i4'),
                ('score', 'f4'),
                ('E', 'f4'),
                ('E_method', 'i4'),
                ('p', 'f4', (1, 3)),
                ('start', 'f4', (1, 3)),
                ('end', 'f4', (1, 3)),
                ('contained', 'bool'),
                ('truth', 'i4', (1, 20)),
                ('truthOverlap', 'f4', (1, 20)),
            ])
            if particle_name in arrakis_file:
                del arrakis_file[particle_name]

            arrakis_file.create_dataset(particle_name, shape=(0,), maxshape=(None,), dtype=particle_data_type)

            """Construct interaction data types"""
            interaction_name = 'standard_record/interaction'
            interaction_data_type = np.dtype([
                ('event_id', 'i4'),
                ('interaction_id', 'i4'),
                ('vtx', 'f4', (1, 3)),
                ('dir_lngtrk', 'f4', (1, 3)),
                ('dir_heshw', 'f4', (1, 3)),
                ('nuhyp', 'f4', (1, 20)),
                ('Enu', 'f4'),
                ('E_method', 'i4'),
                ('part', 'i4', (1, 20)),
                ('truth', 'i4', (1, 20)),
                ('truthOverlap', 'f4', (1, 20)),
            ])
            if interaction_name in arrakis_file:
                del arrakis_file[interaction_name]

            arrakis_file.create_dataset(interaction_name, shape=(0,), maxshape=(None,), dtype=interaction_data_type)

    @profiler
    def distribute_tasks(
        self,
        file_name: str
    ):
        """
        Determine the indices which correspond to unique
        events in the FLOW output files and distribute
        those indices among the workers.

        Args:
            file_name (_str_): _description_
        """
        with h5py.File(self.flow_folder + file_name, 'r', driver='mpio', comm=self.comm) as file:
            if self.rank == 0:
                """Collect event ids from mc_truth and charge/light data"""
                interactions_events = file['mc_truth/interactions/data']['event_id']
                segments_events = file['mc_truth/segments/data']['event_id']
                stack_events = file['mc_truth/stack/data']['event_id']
                trajectories_events = file['mc_truth/trajectories/data']['event_id']
                charge_segments = file['mc_truth/calib_final_hit_backtrack/data']['segment_id']
                non_zero_charge_segments = [row[row != 0] for row in charge_segments]
                max_length = len(max(non_zero_charge_segments, key=len))
                segments_ids = file['mc_truth/segments/data']['segment_id']

                """Determine unique events and distribute among workers"""
                unique_events = np.unique(trajectories_events)
                self.num_events = len(unique_events)
                self.distributed_events.clear()
                self.distributed_interactions_indices.clear()
                self.distributed_segments_indices.clear()
                self.distributed_stack_indices.clear()
                self.distributed_trajectories_indices.clear()
                self.distributed_charge_indices.clear()
                self.distributed_light_indices.clear()

                self.distributed_events = {i: [] for i in range(1, self.size)}
                self.distributed_interactions_indices = {i: [] for i in range(1, self.size)}
                self.distributed_segments_indices = {i: [] for i in range(1, self.size)}
                self.distributed_stack_indices = {i: [] for i in range(1, self.size)}
                self.distributed_trajectories_indices = {i: [] for i in range(1, self.size)}
                self.distributed_charge_indices = {i: [] for i in range(1, self.size)}
                self.distributed_light_indices = {i: [] for i in range(1, self.size)}

                """Determine indices for mc_truth and charge/light data"""
                for i, event_id in enumerate(unique_events):
                    worker_rank = 1 + i % (self.size - 1)  # Distribute round-robin among wor
                    self.distributed_events[worker_rank].append(event_id)
                    self.distributed_interactions_indices[worker_rank].append(
                        np.where(interactions_events == event_id)[0]
                    )
                    self.distributed_segments_indices[worker_rank].append(
                        np.where(segments_events == event_id)[0]
                    )
                    self.distributed_stack_indices[worker_rank].append(
                        np.where(stack_events == event_id)[0]
                    )
                    self.distributed_trajectories_indices[worker_rank].append(
                        np.where(trajectories_events == event_id)[0]
                    )
                    """For charge data we must backtrack through segments"""
                    self.distributed_charge_indices[worker_rank].append(
                        np.any(
                            np.isin(
                                charge_segments[:, :max_length], segments_ids[(segments_events == event_id)]
                            ),
                            axis=1,
                        )
                    )
                    """Likewise for light data, we must backtrack through segments"""
                    self.distributed_light_indices[worker_rank].append([])

                """Distribute indices for events among workers"""
                for ii in range(1, self.comm.size):
                    self.comm.send(self.distributed_events[ii], dest=ii, tag=2)
                    self.comm.send(self.distributed_interactions_indices[ii], dest=ii, tag=3)
                    self.comm.send(self.distributed_segments_indices[ii], dest=ii, tag=4)
                    self.comm.send(self.distributed_stack_indices[ii], dest=ii, tag=5)
                    self.comm.send(self.distributed_trajectories_indices[ii], dest=ii, tag=6)
                    self.comm.send(self.distributed_charge_indices[ii], dest=ii, tag=7)
                    self.comm.send(self.distributed_light_indices[ii], dest=ii, tag=8)
            else:
                self.distributed_events[self.rank] = self.comm.recv(source=0, tag=2)
                self.distributed_interactions_indices[self.rank] = self.comm.recv(source=0, tag=3)
                self.distributed_segments_indices[self.rank] = self.comm.recv(source=0, tag=4)
                self.distributed_stack_indices[self.rank] = self.comm.recv(source=0, tag=5)
                self.distributed_trajectories_indices[self.rank] = self.comm.recv(source=0, tag=6)
                self.distributed_charge_indices[self.rank] = self.comm.recv(source=0, tag=7)
                self.distributed_light_indices[self.rank] = self.comm.recv(source=0, tag=8)

    @profiler
    def process_events_master(
        self,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
    ):
        """
        Special code run by the master node on an event,
        which occurs before any of the workers.  Any work done
        here should be by plugins which create data that is
        needed by the other plugins.

        Args:
            flow_file (_h5py.File_): input flow_file
        """
        pass

    @profiler
    def process_events_worker(
        self,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
    ):
        """
        This function loops over all plugins and hands off the flow_file,
        arrakis_file and the associated indices for each that correspond
        to this workers event.

        Args:
            flow_file (_h5py.File_): input flow_file
        """
        for ii, event in enumerate(self.distributed_events[self.rank]):
            """Grab event index information"""
            event_indices = {
                'interactions': self.distributed_interactions_indices[self.rank][ii],
                'segments': self.distributed_segments_indices[self.rank][ii],
                'stack': self.distributed_stack_indices[self.rank][ii],
                'trajectories': self.distributed_trajectories_indices[self.rank][ii],
                'charge': self.distributed_charge_indices[self.rank][ii],
                'light': self.distributed_light_indices[self.rank][ii]
            }
            event_products = {}
            """Iterate over plugins"""
            for plugin_name, plugin in self.plugins.items():
                plugin.process_event(
                    event=event,
                    flow_file=flow_file,
                    arrakis_file=arrakis_file,
                    event_indices=event_indices,
                    event_products=event_products,
                )

    @profiler
    def run_end_of_file(
        self,
        file_name: str
    ):
        """
        Plugins to be run on the entire file after all
        events have been evaluated within the file.

        One required task is for the master node to gather up all
        reco object information and append it to the arrays in the
        ARRAKIS file.

        Args:
            file_name (_type_): _description_
        """
        pass

    @profiler
    def run_end_of_arrakis(self):
        """
        Set of functions to be run at the end of the entire
        Arrakis job.  Some default operations are to create
        profiling plots.
        """
        self.generate_timing_and_memory_plots()
        if self.rank == 0:
            try:
                self.logger.info("Arrakis program ran successfully. Closing out.")
            except Exception as e:
                self.error_status = e
        else:
            pass

    @profiler
    def generate_timing_and_memory_plots(self):
        """Send over timing and memory information from each worker"""
        if self.rank == 0:
            try:
                collected_timings = {ii: self.comm.recv(source=ii, tag=100) for ii in range(1, self.size)}
                collected_memory = {ii: self.comm.recv(source=ii, tag=101) for ii in range(1, self.size)}
            except Exception as e:
                self.error_status = e
            self.barrier()
        else:
            try:
                self.comm.send(timing_manager.timings, dest=0, tag=100)
                self.comm.send(memory_manager.memory, dest=0, tag=101)
            except Exception as e:
                self.error_status = e
            self.barrier()

        """Generate timing and memory plots"""
        if self.rank == 0:
            try:
                collected_timings[0] = timing_manager.timings
                collected_memory[0] = memory_manager.memory

                timings = {}
                memory = {}
                for key in collected_timings.keys():
                    for name, value in collected_timings[key].items():
                        if name not in timings:
                            timings[name] = value
                            memory[name] = collected_memory[key][name]
                        else:
                            timings[name] += value
                            memory[name] += collected_memory[key][name]

                timing_averages = {}
                timing_stds = {}
                for item in timings.keys():
                    timing_averages[item] = np.mean(timings[item])
                    timing_stds[item] = np.std(timings[item])

                """Separate plugin timings from Arrakis timings"""
                arrakis_fig, arrakis_axs = plt.subplots(figsize=(15, 10))
                arrakis_box_values = []
                arrakis_labels = []
                for item in timings.keys():
                    if 'Plugin' in item:
                        continue
                    else:
                        arrakis_box_values.append(timings[item])
                        arrakis_axs.plot(
                            [],
                            [],
                            marker="",
                            linestyle="-",
                            label=f'{item}\n({timing_averages[item]:.2f} +/- {timing_stds[item]:.2f})',
                        )
                        arrakis_labels.append(item)
                arrakis_axs.boxplot(arrakis_box_values, vert=True, patch_artist=True, labels=arrakis_labels)
                arrakis_axs.set_ylabel(r"$\langle\Delta t\rangle$ (ms)")
                arrakis_axs.set_xticklabels(arrakis_labels, rotation=45, ha="right")
                arrakis_axs.set_yscale("log")
                plt.title(r"Arrakis $\langle\Delta t\rangle$ (ms) vs. function")
                plt.legend(bbox_to_anchor=(1.05, 1.0), loc="upper left")
                plt.tight_layout()
                plt.savefig("/local_scratch/arrakis_nd_timing_avg.png")

                plugin_fig, plugin_axs = plt.subplots(figsize=(15, 10))
                plugin_box_values = []
                plugin_labels = []
                for item in timings.keys():
                    if 'Plugin' in item:
                        plugin_box_values.append(timings[item])
                        plugin_axs.plot(
                            [],
                            [],
                            marker="",
                            linestyle="-",
                            label=f'{item}\n({timing_averages[item]:.2f} +/- {timing_stds[item]:.2f})',
                        )
                        plugin_labels.append(item)
                    else:
                        continue
                plugin_axs.boxplot(plugin_box_values, vert=True, patch_artist=True, labels=plugin_labels)
                plugin_axs.set_ylabel(r"$\langle\Delta t\rangle$ (ms)")
                plugin_axs.set_xticklabels(plugin_labels, rotation=45, ha="right")
                plugin_axs.set_yscale("log")
                plt.title(r"Plugin $\langle\Delta t\rangle$ (ms) vs. function")
                plt.legend(bbox_to_anchor=(1.05, 1.0), loc="upper left")
                plt.tight_layout()
                plt.savefig("/local_scratch/arrakis_nd_plugin_timing_avg.png")

                """Separate plugin memory from Arrakis memory"""
                memory_averages = {}
                memory_stds = {}
                for item in memory.keys():
                    memory_averages[item] = np.mean(memory[item])
                    memory_stds[item] = np.std(memory[item])

                arrakis_fig, arrakis_axs = plt.subplots(figsize=(15, 10))
                arrakis_box_values = []
                arrakis_labels = []
                for item in memory.keys():
                    if 'Plugin' in item:
                        continue
                    else:
                        arrakis_box_values.append(memory[item])
                        arrakis_axs.plot(
                            [],
                            [],
                            marker="",
                            linestyle="-",
                            label=f'{item}\n({memory_averages[item]:.2f} +/- {memory_stds[item]:.2f})',
                        )
                        arrakis_labels.append(item)
                arrakis_axs.boxplot(arrakis_box_values, vert=True, patch_artist=True, labels=arrakis_labels)
                arrakis_axs.set_ylabel(r"$\langle\Delta m\rangle$ (Mb)")
                arrakis_axs.set_xticklabels(arrakis_labels, rotation=45, ha="right")
                arrakis_axs.set_yscale("log")
                plt.title(r"Arrakis $\langle\Delta m\rangle$ (Mb) vs. function")
                plt.legend(bbox_to_anchor=(1.05, 1.0), loc="upper left")
                plt.tight_layout()
                plt.savefig("/local_scratch/arrakis_nd_memory_avg.png")

                plugin_fig, plugin_axs = plt.subplots(figsize=(15, 10))
                plugin_box_values = []
                plugin_labels = []
                for item in memory.keys():
                    if 'Plugin' in item:
                        plugin_box_values.append(memory[item])
                        plugin_axs.plot(
                            [],
                            [],
                            marker="",
                            linestyle="-",
                            label=f'{item}\n({memory_averages[item]:.2f} +/- {memory_stds[item]:.2f})',
                        )
                        plugin_labels.append(item)
                    else:
                        continue
                plugin_axs.boxplot(plugin_box_values, vert=True, patch_artist=True, labels=plugin_labels)
                plugin_axs.set_ylabel(r"$\langle\Delta m\rangle$ (Mb)")
                plugin_axs.set_xticklabels(plugin_labels, rotation=45, ha="right")
                plugin_axs.set_yscale("log")
                plt.title(r"Plugin $\langle\Delta m\rangle$ (Mb) vs. function")
                plt.legend(bbox_to_anchor=(1.05, 1.0), loc="upper left")
                plt.tight_layout()
                plt.savefig("/local_scratch/arrakis_nd_plugin_memory_avg.png")
            except Exception as e:
                self.error_status = e
            self.barrier()
        else:
            self.barrier()

    @profiler
    def run_arrakis_nd(self):
        """
        Main Arrakis program loop.

        Every function call here should be encapsulated in a try/except
        block to prevent hangups with comm.Barrier.

        I. Set up input files for run
        ~ Loop over each file
            a. Load file to determine unique events and indices for each array type
            b. Run whole file over file begin plugins
            c. Set up output arrays in output h5 file
            d. Send index, output information to each worker
            ~ Loop over each event in each worker
                i. Grab event related information and pass to plugins in order
                ii. Collect output information and add to output files
            e. Run whole file over file end plugins
            f. Run end of file functions
        II. Run end of program functions
        """
        self.barrier()

        """Set up input files"""
        try:
            self.set_up_input_files()
        except Exception as e:
            self.error_status = e
        self.barrier()

        """Set up progress bar"""
        if self.rank == 0:
            self.logger.info(f'running Arrakis with {self.size} workers.')
            self.progress_bar = tqdm(
                total=len(self.flow_files),
                ncols=100,
                colour='MAGENTA',
                leave=True,
            )
        self.barrier()

        """Loop over files and call master/worker methods for each."""
        for ii, file_name in enumerate(self.flow_files):
            """First run begin of file"""
            if self.rank == 0:
                try:
                    self.progress_bar.set_description_str(f'File [{ii+1}/{len(self.flow_files)}]')
                    self.run_begin_of_file(file_name)
                except Exception as e:
                    self.error_status = e
            self.barrier()

            """Set up output arrays in ARRAKIS file"""
            if self.rank == 0:
                try:
                    self.set_up_output_arrays(file_name)
                except Exception as e:
                    self.error_status = e
            self.barrier()

            """Prepare indices for workers"""
            try:
                self.distribute_tasks(file_name)
            except Exception as e:
                self.error_status = e
            self.barrier()

            """Now process the file"""
            try:
                arrakis_file_name = file_name.replace('FLOW', 'ARRAKIS').replace('flow', 'arrakis')
                with h5py.File(self.flow_folder + file_name, 'r', driver='mpio', comm=self.comm) as flow_file, \
                     h5py.File(self.arrakis_folder + arrakis_file_name, 'r+', driver='mpio', comm=self.comm) as arrakis_file:
                    self.barrier()
                    if self.rank == 0:
                        """Process event in master node"""
                        try:
                            self.progress_bar.set_postfix_str(f'# Events: [{self.num_events}]')
                            self.process_events_master(flow_file, arrakis_file)
                        except Exception as e:
                            self.error_status = e
                        self.barrier()
                    else:
                        """Process event in worker node"""
                        self.barrier()
                        try:
                            self.process_events_worker(flow_file, arrakis_file)
                        except Exception as e:
                            self.error_status = e
            except Exception as e:
                self.error_status = e
            self.barrier()

            """Run end of file plugins"""
            if self.rank == 0:
                try:
                    self.run_end_of_file(file_name)
                    self.progress_bar.update(1)
                except Exception as e:
                    self.error_status = e
            self.barrier()

        """Run end of program functions"""
        if self.rank == 0:
            self.progress_bar.close()
        self.run_end_of_arrakis()



"""
Logger class for all blip classes.
"""
import warnings
import logging
import platform
import traceback
import socket
import re
import uuid
import psutil
import os


class ArrakisError(Exception):
    """Custom default error for Arrakis"""
    def __init__(self, message="An error occurred"):
        self.message = message
        super().__init__(self.message)


logging_level = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}

logging_output = [
    "console",
    "file",
    "both",
]

warning_list = {
    "deprecation": DeprecationWarning,
    "import": ImportWarning,
    "resource": ResourceWarning,
    "user": UserWarning,
}

error_list = {
    "attribute": AttributeError,
    "arrakis": ArrakisError,
    "index": IndexError,
    "file": FileExistsError,
    "memory": MemoryError,
    "runtime": RuntimeError,
    "type": TypeError,
    "value": ValueError,
}


class LoggingFormatter(logging.Formatter):
    """
    Formatting for Arrakis logging.

    Args:
        logging (_logging.Formatter_): The formatter
        to be edited.

    Returns:
        _logging.Formatter_: The edited formatter
    """
    console_extra = " [%(name)s]: %(message)s"
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    blue = "\x1b[1;34m"
    purple = "\x1b[1;35m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    FORMATS = {
        logging.DEBUG: "[" + grey + "%(levelname)s" + reset + "] [" + purple + "%(name)s" + reset + "]: %(message)s",
        logging.INFO: "[" + blue + "%(levelname)s" + reset + "] [" + purple + "%(name)s" + reset + "]: %(message)s",
        logging.WARNING: "[" + yellow + "%(levelname)s" + reset + "] [" + purple + "%(name)s" + reset + "]: %(message)s",
        logging.ERROR: "[" + red + "%(levelname)s" + reset + "] [" + purple + "%(name)s" + reset + "]: %(message)s",
        logging.CRITICAL: "[" + bold_red + "%(levelname)s" + reset + "] [" + purple + "%(name)s" + reset + "]: %(message)s",
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class Logger:
    """
    Custom logging wrapper for Arrakis programs.
    The logger writes to three different streams:
        (1) console
        (2) file
        (3) debug
    """

    def __init__(
        self,
        name: str = "default",
        level: str = "debug",
        output: str = "file",
        output_file: str = "log",
        file_mode: str = "a",
    ):
        """
        Initializer for the logger

        Args:
            name (str, optional): name for this logger (will appear
            after the log message type [ERROR] [name]). Defaults to "default".

            level (str, optional): _description_. Defaults to "debug".
            output (str, optional): whether to log to console, file
            or both. Defaults to "file".

            output_file (str, optional): name of the output file. Defaults to "log".
            file_mode (str, optional): whether to append, or rewrite
            log files for this run. Defaults to "a".

        Raises:
            ValueError: _description_
            ValueError: _description_
        """
        # check for mistakes
        if level not in logging_level.keys():
            raise ValueError(f"Logging level {level} not in {logging_level}.")
        if output not in logging_output:
            raise ValueError(f"Logging handler {output} not in {logging_output}.")

        # create the logging directory
        if "LOCAL_SCRATCH" in os.environ.keys():
            self.local_log_dir = os.environ["LOCAL_SCRATCH"] + "/.logs/"
        else:
            self.local_log_dir = "/local_scratch/.logs/"
        if not os.path.isdir(self.local_log_dir):
            os.makedirs(self.local_log_dir)

        # use the name as the default output file name
        self.name = name
        self.level = logging_level[level]
        self.output = output
        self.output_file = output_file
        self.file_mode = file_mode

        # create logger
        self.logger = logging.getLogger(self.name)
        self.debug_logger = logging.getLogger(self.name + "_debug")

        # set level
        self.logger.setLevel(self.level)
        self.debug_logger.setLevel(self.level)

        # set format
        self.dateformat = "%H:%M:%S"

        self.console_formatter = LoggingFormatter()
        self.file_formatter = logging.Formatter(
            "[%(asctime)s] [%(levelname)s] [%(name)s]: %(message)s", self.dateformat
        )

        self.debug = logging.FileHandler(
            self.local_log_dir + "/" + self.output_file + ".debug", mode="a"
        )
        self.debug.setLevel(self.level)

        # create handler
        if self.output == "console" or self.output == "both":
            self.console = logging.StreamHandler()
            self.console.setLevel(self.level)
            self.console.setFormatter(self.console_formatter)
            self.logger.addHandler(self.console)
        if self.output == "file" or self.output == "both":
            self.file = logging.FileHandler(
                self.local_log_dir + "/" + self.output_file + ".log", mode="a"
            )
            self.file.setLevel(logging.DEBUG)
            self.file.setFormatter(self.file_formatter)
            self.logger.addHandler(self.file)
        self.debug.setFormatter(self.file)
        self.debug_logger.addHandler(self.debug)
        self.logger.propagate = False

    def info(
        self,
        message: str,
    ):
        """_summary_

        Args:
            message (str): _description_

        Returns:
            _type_: _description_
        """
        """Output to the standard logger "info" """
        return self.logger.info(message)

    def debug(
        self,
        message: str,
    ):
        """_summary_

        Args:
            message (str): _description_

        Returns:
            _type_: _description_
        """
        """Output to the standard logger "debug" """
        return self.debug_logger.debug(message)

    def warn(
        self,
        message: str,
    ):
        """_summary_

        Args:
            message (str): _description_

        Returns:
            _type_: _description_
        """
        """Output to the standard logger "warning" """
        return self.logger.warning(message)

    def warning(
        self,
        message: str,
        warning_type: str = "user",
    ):
        """_summary_

        Args:
            message (str): _description_
            warning_type (str, optional): _description_. Defaults to "user".

        Returns:
            _type_: _description_
        """
        """Output to the standard logger "warning" """
        formatted_lines = traceback.format_stack()[-2]
        if warning_type not in warning_list.keys():
            warning_type = "user"
        self.logger.warning(message)
        warnings.warn(
            f"traceback: {formatted_lines}\nerror: {message}",
            warning_list[warning_type],
        )
        if self.output == "file":
            return self.logger.warning(
                f"traceback: {formatted_lines}\nerror: {message}"
            )
        return

    def error(
        self,
        message: str,
        error_type: str = "arrakis",
    ):
        """_summary_

        Args:
            message (str): _description_
            error_type (str, optional): _description_. Defaults to "value".

        Raises:
            error_list: _description_
        """
        """Output to the standard logger "error" """
        formatted_traceback = ''.join(traceback.format_stack())
        if error_type not in error_list.keys():
            error_type = "arrakis"
        log_message = f"Traceback: \n{formatted_traceback}\nError: {message}"
        if self.output == "file":
            self.logger.error(log_message)
        else:
            self.logger.error(message)
        raise error_list[error_type](log_message)

    def critical(
        self,
        message: str
    ):
        """
        """
        return self.logger.critical(message)

    def get_system_info(
        self
    ) -> dict:
        """
        Attempt to get system info using various
        python packages.  If this fails, return
        an empty dictionary.

        Returns:
            _dict_: dictionary containing system info.
        """
        info = {}
        try:
            info["platform"] = platform.system()
            info["platform-release"] = platform.release()
            info["platform-version"] = platform.version()
            info["architecture"] = platform.machine()
            info["hostname"] = socket.gethostname()
            info["ip-address"] = socket.gethostbyname(socket.gethostname())
            info["mac-address"] = ":".join(re.findall("..", "%012x" % uuid.getnode()))
            info["processor"] = platform.processor()
            info["physical_cores"] = psutil.cpu_count(logical=False)
            info["logical_cores"] = psutil.cpu_count(logical=True)
            info["local_scratch"] = str(round(psutil.disk_usage('/local_scratch').free / (1024.0**3))) + " GB"
            info["local_data"] = str(round(psutil.disk_usage('/local_data').free / (1024.0**3))) + " GB"
            info["RAM"] = str(round(psutil.virtual_memory().total / (1024.0**3))) + " GB"
        except Exception as e:
            self.logger.warning(f"Unable to obtain system information: {e}.")
        return info



"""
"""
from enum import Enum


class Topology(Enum):
    """
    Topological decsriptions of energy
    deposits.
    """
    Undefined = -1
    Noise = 0
    Track = 1
    Shower = 2
    Blip = 3


class PhysicsMicro(Enum):
    """
    micro-level descriptions of topological types.
    This further breaks down the blip/track/shower
    topology labels into micro-level physics.
    """
    Undefined = -1
    Noise = 0
    MIPIonization = 1
    HIPIonization = 2
    ElectronIonization = 3
    Bremsstrahlung = 4
    Annihilation = 5
    PhotoElectric = 6
    GammaCompton = 7
    GammaConversion = 8
    HadronElastic = 9
    HadronInelastic = 10


class PhysicsMacro(Enum):
    """
    Macro-lebel decsriptions of topological types.
    """
    Undefined = -1
    Noise = 0
    MIP = 1
    HIP = 2
    DeltaElectron = 3
    MichelElectron = 4
    ElectronShower = 5
    PositronShower = 6
    PhotonShower = 7
    LowEnergyIonization = 8
    NeutronCaptureGamma474 = 9
    NeutronCaptureGamma336 = 10
    NeutronCaptureGamma256 = 11
    NeutronCaptureGamma118 = 12
    NeutronCaptureGamma083 = 13
    NeutronCaptureGamma051 = 14
    NeutronCaptureGamma016 = 15
    NeutronCaptureGammaOther = 16
    Pi0Decay = 17
    AlphaDecay = 18
    BetaDecay = 19
    GammaDecay = 20
    NuclearRecoil = 21
    ElectronRecoil = 22



"""
"""
from enum import Enum


class ProcessType(Enum):
    """
    Geant4 process identifications.
    """
    Undefined = -1
    NotDefined = 0
    Transportation = 1
    Electromagnetic = 2
    Optical = 3
    Hadronic = 4
    PhotoLeptonHadron = 5
    Decay = 6
    General = 7
    Parameterization = 8
    UserDefined = 9
    Parallel = 10
    Phonon = 11
    UCN = 12


class SubProcessType(Enum):
    """
    Geant4 subprocess identifications.
    """
    Undefined = -1
    Primary = 0
    CoulombScattering = 1
    Ionization = 2
    Bremsstrahlung = 3
    PairProdByCharge = 4
    Annihilation = 5
    AnnihilationToMuMu = 6
    AnnihilationToHadrons = 7
    NuclearStopping = 8
    MultipleScattering = 9
    Rayleigh = 11
    PhotoElectricEffect = 12
    ComptonScattering = 13
    GammaConversion = 14
    GammaConversionToMuMu = 15
    Cerenkov = 21
    Scintillation = 22
    SynchrotronRadiation = 23
    TransitionRadiation = 24
    OpticalAbsorption = 31
    OpticalBoundary = 32
    OpticalRayleigh = 33
    OpticalWLS = 34
    OpticalMieHG = 35
    UCNLoss = 41
    UCNAbsorption = 42
    UCNBoundary = 43
    UCNMultiScattering = 44
    LowEnergyElastic = 51
    LowEnergyExcitation = 52
    LowEnergyIonization = 53
    LowEnergyVibrationalExcitation = 54
    LowEnergyAttachment = 55
    LowEnergyChargeDecrease = 56
    LowEnergyChargeIncrease = 57
    LowEnergyElectronSolvation = 58
    LowEnergyMolecularDecay = 59
    LowEnergyTransportation = 60
    LowEnergyBrownianTransportation = 61
    LowEnergyDoubleIonization = 62
    LowEnergyDoubleCap = 63
    LowEnergyIoniTransfer = 64
    LowEnergyStaticMol = 65
    LowEnergyScavenger = 66
    Transportation = 91
    CoupledTransportation = 92
    HadronElastic = 111
    NeutronGeneral = 116
    HadronInelastic = 121
    HadronCapture = 131
    MuAtomicCapture = 132
    HadronFission = 141
    HadronCaptureAtRest = 151
    LeptonCaptureAtRest = 152
    ChargeExchange = 161
    NuOscillation = 165
    NuElectron = 166
    NuNucleus = 167
    Decay = 201
    DecayRadioactive = 210
    DecayUnknown = 211
    DecayMuAtom = 221
    DecayExternal = 231
    FastSimManagerProcess = 301
    EMDissociation = 310
    StepLimiter = 401
    UserSpecialCuts = 402
    NeutronKiller = 403
    ParallelWorld = 491
    DNAUnknownModel = 11000
    Ritchie1994eSolvation = 11001
    Terrisol1990eSolvation = 11002
    Meesungnoen2002eSolvation = 11003
    Kreipl2009eSolvation = 11004
    Meesungnoensolid2002eSolvation = 11005



"""
Utilities for ArrakisND
"""
from collections import defaultdict
import functools
from datetime import datetime
import psutil
import time
import os


def get_datetime():
    """
    Returns a datetime in the format of:
    YYYY.MM.DD.HH.MM.SS

    Returns:
        _type_: _description_
    """
    time = datetime.now()
    now = f"{time.year}.{time.month}.{time.day}.{time.hour}.{time.minute}.{time.second}"
    return now


class ResetableIterator:
    """_summary_
    """
    def __init__(self):
        self._value = 0

    def __iter__(self):
        return self

    def __next__(self):
        self._value += 1
        return self._value

    def __call__(self):
        return self.__next__()

    def reset(self):
        self._value = 0


def remove_sublist(original_list, remove_list):
    """_summary_

    Args:
        original_list (_type_): _description_
        remove_list (_type_): _description_

    Returns:
        _type_: _description_
    """
    return list(filter(lambda x: x not in remove_list, original_list))


class TimingManager:
    """_summary_
    """
    def __init__(self):
        self.timings = defaultdict(list)

    def record_timing(self, func_name, elapsed):
        self.timings[func_name].append(elapsed)


class MemoryManager:
    """_summary_
    """
    def __init__(self):
        self.memory = defaultdict(list)

    def record_memory(self, func_name, elapsed):
        self.memory[func_name].append(elapsed)


# Global instance to store timings
timing_manager = TimingManager()
# Global instance to store profiles
memory_manager = MemoryManager()


def profiler(func):
    """_summary_

    Args:
        func (_type_): _description_

    Returns:
        _type_: _description_
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        process = psutil.Process(os.getpid())
        mem_before = process.memory_info().rss / 1024 ** 2  # Convert bytes to MB

        start_time = time.time()
        result = func(*args, **kwargs)

        end_time = time.time()
        elapsed_time = end_time - start_time

        mem_after = process.memory_info().rss / 1024 ** 2  # Convert bytes to MB
        mem_used = mem_after - mem_before

        # Capture the memory profile output
        try:
            memory_manager.record_memory(func.__qualname__, mem_used)
            timing_manager.record_timing(func.__qualname__, elapsed_time)
        except Exception:
            memory_manager.record_memory(func.__name__, mem_used)
            timing_manager.record_timing(func.__name__, elapsed_time)

        return result
    return wrapper



"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin


class DaughterPlugin(Plugin):
    """
    A plugin for constructing daughter arrays for event trajectories.

    The FLOW files do not contain daughter information, most likely because
    it would need to exist either as a jagged array, which is not the best
    option for h5 files, or as a large array possibly 100's of columns in
    order to account for trajectories which have a large number of daughters.
    This is similar to what happens with the hit -> segment ID associations
    which use a 200 wide array in order to accomodate hits which have a large
    number of contributing segments in truth.  Most of the entries are zero
    however, which is also the case with daughters.  We therefore have to
    construct the list of daughters on the fly, which is time consuming, but
    otherwise necessary for other parts of the logic.

    In MiniRun4.5 there was a change to the traj_id label, which was converted
    back to the label from edep-sim, so that each traj_id no longer resets in a
    file, but resets per interaction.  Because of this, finding associations
    between traj_id, parent_id and now daughters must include a match to
    vertex_id.  The old way to do this would have been:

        traj_ids = trajectories['traj_id']
        parent_ids = trajectories['parent_id']

        for ii, traj_id in enumerate(traj_ids):
            indices = np.where(parent_ids == traj_id)[0]
            # Store these indices in the dictionary
            daughters[ii] = list(indices)

    but is now changed to reflect the new logic.  This is confirmed to work with the
    old <4.5 miniruns as well.

    The above logic is quite slow (repeated calls to np.where).  In order to
    optimize this, we opt for this better solution:

        unique_pairs = set(zip(parent_ids, vertex_ids))
        pair_to_indices = {pair: [] for pair in unique_pairs}

        for index, pair in enumerate(zip(parent_ids, vertex_ids)):
            pair_to_indices[pair].append(index)

        for ii, (traj_id, vertex_id) in enumerate(zip(traj_ids, vertex_ids)):
            # Direct lookup instead of np.where
            daughters[ii] = pair_to_indices.get((traj_id, vertex_id), [])

        event_products['daughters'] = daughters

    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(DaughterPlugin, self).__init__(config)

        self.input_products = None
        self.output_product = 'daughters'

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        daughters = [[] for ii in range(len(trajectories))]

        traj_ids = trajectories['traj_id']
        vertex_ids = trajectories['vertex_id']
        parent_ids = trajectories['parent_id']

        unique_pairs = set(zip(parent_ids, vertex_ids))
        pair_to_indices = {pair: [] for pair in unique_pairs}

        for index, pair in enumerate(zip(parent_ids, vertex_ids)):
            pair_to_indices[pair].append(index)

        for ii, (traj_id, vertex_id) in enumerate(zip(traj_ids, vertex_ids)):
            # Direct lookup instead of np.where
            daughters[ii] = pair_to_indices.get((traj_id, vertex_id), [])

        event_products['daughters'] = daughters



# interactive_dashboard.py

import panel as pn
import holoviews as hv
import numpy as np

# Ensure compatibility for running in notebooks
hv.extension('bokeh')
pn.extension()

# Define your interactive function
def interactive_scatter(n_points=100):
    x = np.random.randn(n_points)
    y = np.random.randn(n_points)
    return hv.Scatter((x, y)).opts(size=5, color='blue')

# Create a slider widget
slider = pn.widgets.IntSlider(name='Number of points', start=10, end=1000, step=10, value=100)

# Bind the function and slider
scatter_plot = pn.bind(interactive_scatter, n_points=slider)

# Define the layout
layout = pn.Column(slider, scatter_plot)


"""
Script for generating hyperparameter config files from a base config
"""
import os
import numpy as np
import argparse
import csv
import glob

from arrakis_nd.utils.logger import Logger
from arrakis_nd.utils.config import ConfigParser


def run():
    parser = argparse.ArgumentParser(
        prog='ArrakisND Config Generator',
        description='This program constructs a ...',
        epilog='...'
    )
    parser.add_argument(
        'config_file', metavar='<str>.yml', type=str,
        help='config file specification for a module.'
    )
    parser.add_argument(
        '-arrakis_config_location', dest='arrakis_config_location', default='/local_blip',
        help='location for the local blip directory.'
    )
    parser.add_argument(
        '-number_of_processes', dest='number_of_processes', default=1,
        help='number of processes for arrakis runs.'
    )

    logger = Logger('arrakis_generator', output="both", file_mode="w")

    args = parser.parse_args()
    config_file = args.config_file
    arrakis_config_location = args.arrakis_config_location
    number_of_processes = int(args.number_of_processes)

    config_parser = ConfigParser(config_file)
    config = config_parser.data

    arrakis_dict = config["arrakis_nd"]
    simulation_folder = arrakis_dict["simulation_folder"]
    simulation_files = arrakis_dict["simulation_files"]
    if isinstance(arrakis_dict["simulation_files"], list):
        simulation_files = [
            input_file for input_file in arrakis_dict["simulation_files"]
            if input_file not in arrakis_dict["skip_files"]
        ]
    elif isinstance(arrakis_dict["simulation_files"], str):
        if arrakis_dict["simulation_files"] == "all":
            logger.info(
                f"searching {simulation_folder} recursively for all .h5 files."
            )
            simulation_files = [
                os.path.basename(input_file) for input_file in glob.glob(
                    f"{simulation_folder}*.h5", recursive=True
                )
                if input_file not in arrakis_dict["skip_files"]
            ]
        else:
            try:
                logger.info(
                    f'searching {simulation_folder} recursively for all {arrakis_dict["simulation_files"]} files.'
                )
                simulation_files = [
                    os.path.basename(input_file) for input_file in glob.glob(
                        f'{simulation_folder}{arrakis_dict["simulation_files"]}',
                        recursive=True,
                    )
                    if input_file not in arrakis_dict["skip_files"]
                ]
            except Exception as exception:
                logger.error(
                    f'specified "simulation_files" parameter: {arrakis_dict["simulation_files"]} incompatible!'
                    + f" exception: {exception}"
                )
    else:
        logger.error(
            f'specified "simulation_files" parameter: {arrakis_dict["simulation_files"]} incompatible!'
        )
    # split up simulation_files
    file_chunk_size = int(np.ceil(len(simulation_files) / number_of_processes))
    file_chunks = [simulation_files[i:i+file_chunk_size] for i in range(0, len(simulation_files), file_chunk_size)]

    arrakis_folders = []

    for ii in range(len(file_chunks)):
        random_config = config.copy()
        random_config['arrakis_nd']['simulation_files'] = file_chunks[ii]
        if not os.path.isdir(f'{os.path.abspath(arrakis_config_location)}/arrakis_config_iteration_{ii}'):
            os.makedirs(f'{os.path.abspath(arrakis_config_location)}/arrakis_config_iteration_{ii}')
        arrakis_folders.append([f'{os.path.abspath(arrakis_config_location)}/arrakis_config_iteration_{ii}'])
        config_parser.save_config(
            random_config,
            f'{os.path.abspath(arrakis_config_location)}/arrakis_config_iteration_{ii}/arrakis_config.yaml'
        )
    with open(f'{os.path.abspath(arrakis_config_location)}/arrakis_data.csv', "w") as file:
        writer = csv.writer(file, delimiter=",")
        writer.writerows(arrakis_folders)


if __name__ == "__main__":
    run()



"""
ArrakisND main program
"""
import argparse
from mpi4py import MPI

from arrakis_nd.utils.logger import Logger
from arrakis_nd.utils.config import ConfigParser
from arrakis_nd.arrakis.arrakis import Arrakis


def run():
    """_summary_
    """

    """
    We do a preliminary check to ensure that MPI is available
    and that the number of processes is at least 2.  Assuming
    that ArrakisND is being run on a system with more than one
    CPU core.
    """
    logger = Logger("arrakis_runner", output="both")

    try:
        comm = MPI.COMM_WORLD
        size = comm.Get_size()

        # check that size >= 2, which is required
        # for Arrakis to run.  Otherwise we quit early
        if size < 2:
            logger.error(f"Number of processes must be >=2! Received {size} from MPI", "value")
    except Exception as e:
        logger.error(f"Error occurred with gathering MPI: {e}", "runtime")

    parser = argparse.ArgumentParser(
        prog="Arrakis Module Runner",
        description="This program runs the Arrakis module " + "from a config file.",
        epilog="...",
    )
    parser.add_argument(
        "config_file",
        metavar="<config_file>.yml",
        type=str,
        help="config file specification for a BLIP module.",
    )
    parser.add_argument(
        "-name",
        dest="name",
        default="arrakis",
        help='name for this run (default "arrakis").',
    )
    parser.add_argument(
        "-n",
        dest="number_of_files",
        default=None,
        help='number of files to process (default None).',
    )
    args = parser.parse_args()

    config_file = args.config_file
    name = args.name
    number_of_files = args.number_of_files

    try:
        config = ConfigParser(config_file).data
    except Exception as e:
        logger.error(f"failed to parse config: {e}")

    if number_of_files is not None:
        try:
            number_of_files = int(number_of_files)
        except Exception as e:
            logger.error(f"unable to convert number_of_files from {type(number_of_files)} to int: {e}")

        if isinstance(number_of_files, int):
            if number_of_files > 0:
                config["arrakis_nd"]["number_of_files"] = number_of_files
    meta = {
        "name": name,
        "config_file": config_file
    }

    try:
        arrakis = Arrakis(config, meta)
    except Exception as e:
        logger.error(f"failed to construct Arrakis object: {e}")

    arrakis.run_arrakis_nd()


if __name__ == "__main__":
    run()



"""
"""
import h5py


class Plugin:
    """
    A base class for plugins to Arrakis.

    Plugins are classes which process flow data and produce
    either intermediate data products or arrakis output.
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        All plugins must run super(CustomPlugin, self).__init__(config)
        and specify any required data_products and any output data_products
        so that Arrakis knows they are called in the correct sequence.

        Args:
            config (dict, optional): _description_. Defaults to {}.
        """
        self.config = config

        """These should be set by the creator of the custom plugin"""
        self.input_products = None
        self.output_product = None

    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        Event processing function for a plugin. The process_event
        function requires both a flow file and arrakis file, as
        well as the dictionary of event_indices corresponding to
        both.  The event_products is a shared dictionary which
        can be updated with intermmediate data used by other
        plugins.


        This function must be overriden by the inherited class.

        Args:
            flow_file (h5py.File): _description_
            arrakis_file (h5py.File): _description_
            event_indices (dict): _description_
            event_products (dict): _description_
        """
        raise NotImplementedError("Plugin classes must implement this function!")



"""
"""
import h5py
import numpy as np
from numba import jit

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin


class TrackIDHitMapPlugin(Plugin):
    """
    A plugin for ...
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        This plugin generates a map from (traj_id, vertex_it) pairs to
        indices in the calib_final_hits array.  This allows other plugins
        to easily find which hits originate from each particle so that
        labels can be assigned in the corresponding ARRAKIS array.

        There are several approaches to associating (traj_id, vertex_id) to hits.  The
        slowest way is to iterate over the (traj_id, vertex_id) pairs in the
        trajectories array and try to see if any hits correspond to those ids.
        This can be done like this:

            for ii, (traj_id, vertex_id) in enumerate(
                zip(trajectories_traj_ids, trajectories_vertex_ids)
            ):
                segment_indices = np.where(
                    (segments_traj_ids == traj_id) & (segments_vertex_ids == vertex_id)
                )[0]
                segment_ids = segments_segment_ids[segment_indices]
                charge_ids = np.any(
                    np.isin(
                        charge_segment_ids, segment_ids
                    ),
                    axis=1,
                )

        This is unfortunately very slow, so instead we opt for the solution
        which is currently implemented in process_event.

        """
        super(TrackIDHitMapPlugin, self).__init__(config)

        self.input_products = None
        self.output_product = 'track_id_hit_map'

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        Here we associate (traj_id, vertex_id) pairs to indices in the event_indices['charge']
        mask so that later plugins can easily grab the hits associated to each particle.
        To do this we must first associate (traj_id, vertex_id) to (segment_id) and then
        use the back tracking information in 'mc_truth/calib_final_hit_backtrack/data'
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        segments = flow_file['mc_truth/segments/data'][event_indices['segments']]
        charge_back_track = flow_file['mc_truth/calib_final_hit_backtrack/data'][event_indices['charge']]

        trajectories_traj_ids = trajectories['traj_id']
        trajectories_vertex_ids = trajectories['vertex_id']
        segments_traj_ids = segments['traj_id']
        segments_vertex_ids = segments['vertex_id']
        segments_segment_ids = segments['segment_id']
        charge_segment_ids = charge_back_track['segment_id']

        """Create empty map with (traj_id, vertex_id) as keys"""
        track_id_hit_map = {
            (traj_id, vertex_id): []
            for (traj_id, vertex_id) in zip(trajectories_traj_ids, trajectories_vertex_ids)
        }

        """Loop over segment ids"""
        for ii, segment_ids in enumerate(charge_segment_ids):
            """Get segment_ids of segments of each hit"""
            segment_ids = segment_ids[(segment_ids != 0)]
            """Determine where these ids are in the segments array"""
            segment_indices = np.array([
                np.where(segments_segment_ids == segment_id)[0][0]
                if segment_id in segments_segment_ids else -1
                for segment_id in segment_ids
            ])
            """Associate this hit to the (traj_id, vertex_id) pairs"""
            for segment_index in segment_indices:
                track_id_hit_map[
                    (segments_traj_ids[segment_index], segments_vertex_ids[segment_index])
                ].append(ii)

        event_products['track_id_hit_map'] = track_id_hit_map



"""
Tools for parsing config files
"""
import yaml

from arrakis_nd.utils.logger import Logger
from arrakis_nd.utils.utils import profiler


class ConfigParser:
    """
    Config file parser for Arrakis.  This class simply
    wraps the pyyaml class with some added functionality,
    such as being able to load nested configs which refer
    to each other.  This can be done by putting

        load_config:    "config_file.yaml"

    in the input config file.  The parameters will be chosen
    in a hierarchical manner (i.e., the parameters can be
    overwritten by respecifying them in higher up configs)
    """

    @profiler
    def __init__(
        self,
        config_file: str,
    ):
        """
        Config file parser for Arrakis initializer

        Args:
            config_file (str): The location of the config
            file to be loaded.
        """
        self.config_file = config_file
        self.nested_config_files = [config_file]
        self.nested_configs = []
        self.data = {}
        self.parse_config()

    @profiler
    def parse_config(self):
        """
        Main function of the config parser.
        """
        self.collect_nested_configs(self.config_file)
        self.nested_config_files.reverse()
        self.nested_configs.reverse()
        for config in self.nested_configs:
            self.data.update(config)

    @profiler
    def collect_nested_configs(
        self,
        config_file: str,
    ):
        """
        Load configs in a hierarchical manner, replacing
        parameters with ones which occur at higher levels.

        Args:
            config_file (_str_): The location of the config
            file to be loaded.
        """
        with open(config_file, 'r') as file:
            temp_data = yaml.safe_load(file)
            self.nested_configs.append(temp_data)
        if 'load_config' in temp_data.keys():
            self.nested_config_files.append(temp_data['load_config'])
            self.collect_nested_configs(temp_data['load_config'])

    @profiler
    def save_config(
        self,
        config_dictionary: str,
        output_file: str,
    ):
        """
        Save a dictionary to a config file.

        Args:
            config_dictionary (_str_): The location of the config
            file to be loaded.
            output_file (_str_): The location and name of the output
            yaml file.
        """
        try:
            with open(output_file, 'w') as file:
                yaml.dump(config_dictionary, file)
        except Exception as e:
            raise RuntimeError(f"Failed to save config dictionary {config_dictionary} to output file {output_file}: {e}")



"""
This is an empty plugin that can be used as a template.
There are several comments throughout which provide detail
on the functions which must be overidden (user_name is replaced
with the $USER from the sesseion, but you should change it
to whatever you want).

name:   [*user_name*]
date:   [*user_date*]

"""
import h5py

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin


class EmptyPlugin(Plugin):
    """
    This custom plugin inherits from the base class
    Plugin, which is used in the main event loop of the
    Arrakis program.  Plugins are called by every worker
    that is processing events, and are called on every event.

    For each event, a pointer to the FLOW and ARRAKIS files
    are passed to the "process_event" function below.

    A dictionary of indices which point to the FLOW and ARRAKIS
    files is also passed as "event_indices" which contain the
    following keys:

    event_indices = {
        'interactions': indices of the "event" in the 'mc_truth/interactions/data' array
        'segments':     indices of the "event" in the 'mc_truth/segments/data' array
        'stack':        indices of the "event" in the 'mc_truth/stack/data' array
        'trajectories': indices of the "event" in the 'mc_truth/trajectories/data' array
        'charge':       indices of the "event" in the 'charge/calib_final_hits/data' array
    }
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        To keep things as simple as possible, we restrict plugins to
        have only a single input parameter "config", which is a dictionary
        that can contain any number of parameters that can be specified
        at run-time in the corresponding config file.

        The 'input_products' and 'output_produces' must be configured!
        The 'input_products' should be either 'None', 'str' or a list of 'str'.  The
        'str's tell Arrakis what plugin output_products that this plugin needs in
        order to work properly.  For example, the 'daughter_plugin' creates the 
        'daughters' output_product, which contains the indices of the daughters
        for each traj_id in trajectories.  To specify that this plugin needs
        that product, we would write,

            self.input_products = 'daughters'   -  or  -  self.input_products = ['daughters']

        The 'output_product' should give the name of the intermediate product (if any)
        that this plugin produces and appends to the "event_products" dictionary.
        """
        super(EmptyPlugin, self).__init__(config)

        self.input_products = None
        self.output_product = None

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        pass



"""
Script for generating a plugin template with an accompanying config file
"""
import os
import shutil
import argparse
import socket

from arrakis_nd.utils.utils import get_datetime


def run():
    parser = argparse.ArgumentParser(
        prog='Arrakis Plugin Creator',
        description='This program constructs a Arrakis Plugin ' +
                    'with an empty template.',
        epilog='...'
    )
    parser.add_argument(
        '-plugin_name', dest='plugin_name', default='my',
        help='name to use for the custom plugin parts.'
    )
    parser.add_argument(
        '-plugin_location', dest='plugin_location', default='/local_scratch',
        help='location for the local scratch directory.'
    )

    args = parser.parse_args()
    if args.plugin_location is not None:
        if not os.path.isdir(args.plugin_location):
            args.plugin_location = './'
    else:
        args.plugin_location = './'
    snake_case_name = args.plugin_name
    plugin_name_words = snake_case_name.split('_')
    camel_case_name = ''.join(word.capitalize() for word in plugin_name_words)
    plugin_location = args.plugin_location + '/' + camel_case_name + 'CustomPlugin/'

    # now copy template files and replace 'Empty' with args.plugin_name
    now = get_datetime()
    if not os.path.isdir(plugin_location):
        os.makedirs(plugin_location)
    else:
        if not os.path.isdir(f"{plugin_location}.backup/"):
            os.makedirs(f"{plugin_location}.backup/")
        os.makedirs(f"{plugin_location}.backup/{now}")
        selected_files = [
            file for file in os.listdir(plugin_location)
            if (file.endswith('.py') or file.endswith('.yaml'))
        ]
        for file in selected_files:
            source_path = os.path.join(plugin_location, file)
            destination_path = os.path.join(f"{plugin_location}.backup/{now}", file)
            shutil.move(source_path, destination_path)

    # add *user_name* and *user_date*
    try:
        user_name = socket.gethostname()
    except Exception:
        user_name = "*user_name*"

    shutil.copy(
        os.path.dirname(__file__) + '/../plugins/empty_plugin.py',
        f'{plugin_location}/{snake_case_name}.py'
    )
    with open(f'{plugin_location}/{snake_case_name}.py', 'r') as file:
        content = file.read()
    modified_content = content.replace(
        'Empty',
        camel_case_name
    )
    modified_content = modified_content.replace(
        'empty',
        snake_case_name
    )
    modified_content = modified_content.replace(
        "*user_name*", user_name
    )
    modified_content = modified_content.replace(
        "*user_date*", now
    )
    with open(f'{plugin_location}{snake_case_name}.py', 'w') as file:
        file.write(modified_content)


if __name__ == "__main__":
    run()



from setuptools import find_packages
from setuptools import setup

with open("README.md", "r") as file:
    long_description = file.read()

# Function to read the list of requirements from requirements.txt
def read_requirements():
    with open('requirements.txt') as req:
        return req.read().splitlines()

setup(
    # name
    name="arrakis_nd",
    # current version
    #   MAJOR VERSION:  02
    #   MINOR VERSION:  00
    #   Maintenance:    00
    version="2.00.00",
    # descriptions
    description="Arrakis module for near detector data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="",
    # my info
    author="Nicholas Carrara, Marjolein van Nuland, Luis Lepin",
    author_email="ncarrara.physics@gmail.com",
    # where to find the source
    url="https://github.com/Neutron-Calibration-in-DUNE/ArrakisND",
    # requirements
    install_requires=read_requirements(),
    # packages
    packages=find_packages(
        # where='pointnet',
        exclude=["test"],
    ),
    include_package_data=True,
    # classifiers
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">3.7",
    # possible entry point
    entry_points={
        "console_scripts": [
            "arrakis_nd = arrakis_nd.programs.run_arrakis:run",
            "create_arrakis_runs = arrakis_nd.programs.create_arrakis_runs:run",
            'create_plugin = arrakis_nd.programs.plugin_creator:run',
        ],
    },
)
